"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
class Repository {
    constructor(config = {}) {
        this.data = {};
        const { hooks = {}, shouldDeepClone, } = config;
        this.hooks = hooks;
        this.config = config;
    }
    get length() {
        return Object.keys(this.data).length;
    }
    set(key, data, interpolator) {
        if (Array.isArray(key)) {
            return this.multiSet(key, Array.from(data || []));
        }
        const beforeHook = this.hooks.beforeSet;
        if (beforeHook) {
            beforeHook(this, key, data, interpolator);
        }
        if (this.config.shouldDeepClone) {
            data = _.clone(data);
        }
        let returnData = (this.data[key] = data);
        returnData = interpolator ? interpolator(returnData) : returnData;
        const afterHook = this.hooks.afterSet;
        if (afterHook) {
            afterHook(this, key, data, interpolator, returnData);
        }
        return returnData;
    }
    get(key, defaultValue, interpolator) {
        if (Array.isArray(key)) {
            return this.multiGet(key, Array.from(defaultValue || []));
        }
        const beforeHook = this.hooks.beforeGet;
        if (beforeHook) {
            beforeHook(this, key, interpolator);
        }
        let returnData = interpolator
            ? interpolator(this.data[key] || defaultValue)
            : this.data[key] || defaultValue;
        if (this.config.shouldDeepClone) {
            returnData = _.clone(returnData);
        }
        const afterHook = this.hooks.afterGet;
        if (afterHook) {
            afterHook(this, key, defaultValue, interpolator, returnData);
        }
        return returnData;
    }
    remove(key, interpolator) {
        if (Array.isArray(key)) {
            return this.multiRemove(key);
        }
        const beforeHook = this.hooks.beforeRemove;
        if (beforeHook) {
            beforeHook(this, key, interpolator);
        }
        let returnData = this.data[key];
        delete this.data[key];
        if (this.config.shouldDeepClone) {
            returnData = _.clone(returnData);
        }
        const afterHook = this.hooks.afterRemove;
        if (afterHook) {
            afterHook(this, key, interpolator, returnData);
        }
        return returnData;
    }
    clear(interpolator) {
        const beforeHook = this.hooks.beforeClear;
        if (beforeHook) {
            beforeHook(this, interpolator);
        }
        let returnData = this.data;
        this.data = {};
        if (this.config.shouldDeepClone) {
            returnData = _.clone(returnData);
        }
        const afterHook = this.hooks.afterClear;
        if (afterHook) {
            afterHook(this, interpolator, returnData);
        }
        return returnData;
    }
    keys(interpolator) {
        const beforeHook = this.hooks.beforeKeys;
        if (beforeHook) {
            beforeHook(this, interpolator);
        }
        const returnData = interpolator
            ? interpolator(Object.keys(this.data))
            : Object.keys(this.data);
        const afterHook = this.hooks.afterKeys;
        if (afterHook) {
            afterHook(this, interpolator, returnData);
        }
        return returnData;
    }
    values(interpolator) {
        const beforeHook = this.hooks.beforeValues;
        if (beforeHook) {
            beforeHook(this, interpolator);
        }
        const returnData = interpolator
            ? interpolator(Object.values(this.data))
            : Object.values(this.data);
        const afterHook = this.hooks.afterValues;
        if (afterHook) {
            afterHook(this, interpolator, returnData);
        }
        return returnData;
    }
    all(interpolator) {
        const beforeHook = this.hooks.beforeAll;
        if (beforeHook) {
            beforeHook(this, interpolator);
        }
        let returnData = this.data;
        if (this.config.shouldDeepClone) {
            returnData = _.clone(returnData);
        }
        const afterHook = this.hooks.afterAll;
        if (afterHook) {
            afterHook(this, interpolator, returnData);
        }
        return returnData;
    }
    multiRemove(key, interpolator) {
        const beforeHook = this.hooks.beforeMultiRemove;
        if (beforeHook) {
            beforeHook(this, key, interpolator);
        }
        const returnData = key.map(k => this.remove(k));
        const afterHook = this.hooks.afterMultiRemove;
        if (afterHook) {
            afterHook(this, key, interpolator, returnData);
        }
        return returnData;
    }
    multiGet(key, defaultValue, interpolator) {
        const beforeHook = this.hooks.beforeMultiGet;
        if (beforeHook) {
            beforeHook(this, key, defaultValue, interpolator);
        }
        const returnData = key.map((k, i) => this.get(k, defaultValue[i]));
        const afterHook = this.hooks.afterMultiGet;
        if (afterHook) {
            afterHook(this, key, defaultValue, interpolator, returnData);
        }
        return returnData;
    }
    multiSet(key, data, interpolator) {
        const beforeHook = this.hooks.beforeMultiSet;
        if (beforeHook) {
            beforeHook(this, key, data, interpolator);
        }
        const returnData = key.map((k, i) => this.set(k, data[i]));
        const afterHook = this.hooks.afterMultiSet;
        if (afterHook) {
            afterHook(this, key, data, interpolator, returnData);
        }
        return returnData;
    }
}
exports.default = Repository;
//# sourceMappingURL=Repository.js.map